// Corrector plugin for Avisynth
// Copyright(c)2005 Manao as function of MVTools
// Copyright(c)2006 Fizick as standalone plugin, other colorspace formats

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA, or visit
// http://www.gnu.org/copyleft/gpl.html .

#define WIN32_LEAN_AND_MEAN

#include "avisynth.h"
#include "corrector.h"

AVSValue __cdecl Create_Corrector(AVSValue args, void* user_data, IScriptEnvironment* env)
{
	return new Corrector(
		args[0].AsClip(),
      args[1],
      args[2].AsInt(0),
      args[3].AsInt(10),
		env);
}


extern "C" __declspec(dllexport) const char* __stdcall
AvisynthPluginInit2(IScriptEnvironment* env) {

   env->AddFunction("Corrector", "cc+[mode]i[th]i", Create_Corrector, 0);
	return("Corrector");
}

Corrector::Corrector(PClip _child, AVSValue clips, int mode, int th, IScriptEnvironment *env) :
GenericVideoFilter(_child)
{
   if ( !(clips.ArraySize() & 1) )
      env->ThrowError("Corrector : there is not the same number of analysis and output clips");

   nClips = clips.ArraySize() / 2;
   AClip = _child;
   OClip = clips[0].AsClip();

   for ( int i = 0; i < nClips; i ++ )
   {
      AClips.push_back(clips[i+1].AsClip());
      OClips.push_back(clips[i+nClips+1].AsClip());
   }

   nWidth = vi.width;
   nHeight = vi.height;

   if ( mode == 1)
      Process = ProcessBest;
   else
      Process = ProcessMean;

   nThreshold = th;
}

Corrector::~Corrector()
{
}

void ProcessMean(BYTE *pDst, const BYTE *pASrc, const BYTE *pOSrc, const BYTE **pASrcs, const BYTE **pOSrcs, 
                 int nDstPitch, int nASrcPitch, int nOSrcPitch, int *pAPitches, int *pOPitches, int width, int height, int nThreshold, int nClips)
{
   int nSum;
   int nCount;
   for ( int y = 0; y < height; y++ )
   {
      for ( int x = 0; x < width; x++ )
      {
         nCount = nSum = 0;
         for ( int i = 0; i < nClips; i++ )
         {
            if ( abs(pASrcs[i][x] - pASrc[x]) < nThreshold )
            {
               nSum += pOSrcs[i][x];
               nCount++;
            }
         }
         if ( nCount )
            pDst[x] = (nSum + (nCount >> 1)) / nCount;
         else
            pDst[x] = pOSrc[x];
      }
      for ( int i = 0; i < nClips; i++ )
      {
         pASrcs[i] += pAPitches[i];
         pOSrcs[i] += pOPitches[i];
      }
      pASrc += nASrcPitch;
      pOSrc += nOSrcPitch;
      pDst += nDstPitch;
   }
}

void ProcessBest(BYTE *pDst, const BYTE *pASrc, const BYTE *pOSrc, const BYTE **pASrcs, const BYTE **pOSrcs, 
                 int nDstPitch, int nASrcPitch, int nOSrcPitch, int *pAPitches, int *pOPitches, int width, int height, int nThreshold, int nClips)
{
   int nMin;
   int nVal;
   for ( int y = 0; y < height; y++ )
   {
      for ( int x = 0; x < width; x++ )
      {
         nMin = nThreshold;
         nVal = -1;
         for ( int i = 0; i < nClips; i++ )
         {
            if ( abs(pASrcs[i][x] - pASrc[x]) < nMin )
            {
               nVal = pOSrcs[i][x];
               nMin = abs(pASrcs[i][x] - pASrc[x]);
            }
         }
         if ( nVal == -1 )
            pDst[x] = pOSrc[x];
         else
            pDst[x] = nVal;
      }
      for ( int i = 0; i < nClips; i++ )
      {
         pASrcs[i] += pAPitches[i];
         pOSrcs[i] += pOPitches[i];
      }
      pASrc += nASrcPitch;
      pOSrc += nOSrcPitch;
      pDst += nDstPitch;
   }
}


PVideoFrame __stdcall Corrector::GetFrame(int n, IScriptEnvironment* env)
{
   PVideoFrame ASrc = AClip->GetFrame(n, env);
   PVideoFrame OSrc = OClip->GetFrame(n, env);

   PVideoFrame dst = env->NewVideoFrame(vi);
   env->MakeWritable(&dst);

   const BYTE **pASrcs = new const BYTE*[nClips];
   const BYTE **pOSrcs = new const BYTE*[nClips];
   int *pAPitches = new int[nClips];
   int *pOPitches = new int[nClips];

   if(vi.IsYV12())
   {
	   for ( int i = 0; i < nClips; i++ )
	   {
		  pASrcs[i] = (AClips[i]->GetFrame(n, env))->GetReadPtr(PLANAR_Y);
		  pAPitches[i] = (AClips[i]->GetFrame(n, env))->GetPitch(PLANAR_Y);
		  pOSrcs[i] = (OClips[i]->GetFrame(n, env))->GetReadPtr(PLANAR_Y);
		  pOPitches[i] = (OClips[i]->GetFrame(n, env))->GetPitch(PLANAR_Y);
	   }

	   Process((dst)->GetWritePtr(PLANAR_Y), (ASrc)->GetReadPtr(PLANAR_Y), (OSrc)->GetReadPtr(PLANAR_Y), pASrcs, pOSrcs,
			   (dst)->GetPitch(PLANAR_Y), (ASrc)->GetPitch(PLANAR_Y), (OSrc)->GetPitch(PLANAR_Y), pAPitches, pOPitches,
			   nWidth, nHeight, nThreshold, nClips);

	   for ( int i = 0; i < nClips; i++ )
	   {
		  pASrcs[i] = (AClips[i]->GetFrame(n, env))->GetReadPtr(PLANAR_U);
		  pAPitches[i] = (AClips[i]->GetFrame(n, env))->GetPitch(PLANAR_U);
		  pOSrcs[i] = (OClips[i]->GetFrame(n, env))->GetReadPtr(PLANAR_U);
		  pOPitches[i] = (OClips[i]->GetFrame(n, env))->GetPitch(PLANAR_U);
	   }

	   Process((dst)->GetWritePtr(PLANAR_U), (ASrc)->GetReadPtr(PLANAR_U), (OSrc)->GetReadPtr(PLANAR_U), pASrcs, pOSrcs,
			   (dst)->GetPitch(PLANAR_U), (ASrc)->GetPitch(PLANAR_U), (OSrc)->GetPitch(PLANAR_U), pAPitches, pOPitches,
			   (dst)->GetRowSize(PLANAR_U), (dst)->GetHeight(PLANAR_U), nThreshold, nClips);

	   for ( int i = 0; i < nClips; i++ )
	   {
		  pASrcs[i] = (AClips[i]->GetFrame(n, env))->GetReadPtr(PLANAR_V);
		  pAPitches[i] = (AClips[i]->GetFrame(n, env))->GetPitch(PLANAR_V);
		  pOSrcs[i] = (OClips[i]->GetFrame(n, env))->GetReadPtr(PLANAR_V);
		  pOPitches[i] = (OClips[i]->GetFrame(n, env))->GetPitch(PLANAR_V);
	   }

	   Process((dst)->GetWritePtr(PLANAR_V), (ASrc)->GetReadPtr(PLANAR_V), (OSrc)->GetReadPtr(PLANAR_V), pASrcs, pOSrcs,
			   (dst)->GetPitch(PLANAR_V), (ASrc)->GetPitch(PLANAR_V), (OSrc)->GetPitch(PLANAR_V), pAPitches, pOPitches,
			   (dst)->GetRowSize(PLANAR_V), (dst)->GetHeight(PLANAR_V), nThreshold, nClips);
   }
   else
   {
	   for ( int i = 0; i < nClips; i++ )
	   {
		  pASrcs[i] = (AClips[i]->GetFrame(n, env))->GetReadPtr();
		  pAPitches[i] = (AClips[i]->GetFrame(n, env))->GetPitch();
		  pOSrcs[i] = (OClips[i]->GetFrame(n, env))->GetReadPtr();
		  pOPitches[i] = (OClips[i]->GetFrame(n, env))->GetPitch();
	   }

	   Process((dst)->GetWritePtr(), (ASrc)->GetReadPtr(), (OSrc)->GetReadPtr(), pASrcs, pOSrcs,
			   (dst)->GetPitch(), (ASrc)->GetPitch(), (OSrc)->GetPitch(), pAPitches, pOPitches,
			   (dst)->GetRowSize(), (dst)->GetHeight(), nThreshold, nClips);
   }

 
   delete[] pASrcs;
   delete[] pAPitches;
   delete[] pOSrcs;
   delete[] pOPitches;

   return dst;
}